# -*- coding: utf-8 -*-

from .utils import T_DNAMODB_JSON
from .utils import T_DATA
from .import_table import ImportStatusEnum
from .import_table import ImportFormatEnum
from .import_table import ImportJob
from .import_table import write_amazon_ion
from .import_table import write_dynamodb_json
from .export_table import ManifestSummary
from .export_table import DataFile
from .export_table import ExportStatusEnum
from .export_table import ExportFormatEnum
from .export_table import ExportTypeEnum
from .export_table import ExportJob



# -*- coding: utf-8 -*-

"""
DynamoDB import from S3 tool box.

Reference:

- DynamoDB data import from Amazon S3: how it works: https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/S3DataImport.HowItWorks.html
"""

import typing as T
import enum
import gzip
import dataclasses
from datetime import datetime

import botocore.exceptions

from .importer import amazon_ion, dynamodb_json

from .waiter import Waiter
from .utils import split_s3_uri


if T.TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.type_defs import PutObjectOutputTypeDef
    from mypy_boto3_dynamodb.client import DynamoDBClient


class ImportStatusEnum(str, enum.Enum):
    """
    Import job status enumeration.
    """
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    CANCELLING = "CANCELLING"
    CANCELLED = "CANCELLED"
    FAILED = "FAILED"


class ImportFormatEnum(str, enum.Enum):
    """
    Import format enumeration.
    """
    DYNAMODB_JSON = "DYNAMODB_JSON"
    ION = "ION"
    CSV = "CSV"


@dataclasses.dataclass
class ImportJob:
    """
    The DynamoDB import table job data model.

    Ref:

    - import_table: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/import_table.html
    - describe_import: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/describe_import.html
    - list_imports: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/list_imports.html
    - How it works: https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/S3DataImport.HowItWorks.html
    """

    # fmt: off
    arn: str = dataclasses.field()
    status: str = dataclasses.field()
    table_arn: T.Optional[str] = dataclasses.field(default=None)
    table_id: T.Optional[str] = dataclasses.field(default=None)
    client_token: T.Optional[str] = dataclasses.field(default=None)
    s3_bucket_owner: T.Optional[str] = dataclasses.field(default=None)
    s3_bucket: T.Optional[str] = dataclasses.field(default=None)
    s3_prefix: T.Optional[str] = dataclasses.field(default=None)
    error_count: T.Optional[int] = dataclasses.field(default=None)
    cloudwatch_log_group_arn: T.Optional[str] = dataclasses.field(default=None)
    input_format: T.Optional[str] = dataclasses.field(default=None)
    import_format_options: T.Optional[T.Dict[str, str]] = dataclasses.field(default=None)
    import_compression_type: T.Optional[str] = dataclasses.field(default=None)
    table_creation_parameters: T.Optional[T.Dict[str, str]] = dataclasses.field(default=None)
    start_time: T.Optional[datetime] = dataclasses.field(default=None)
    end_time: T.Optional[datetime] = dataclasses.field(default=None)
    processed_size_bytes: T.Optional[int] = dataclasses.field(default=None)
    processed_item_count: T.Optional[int] = dataclasses.field(default=None)
    imported_item_count: T.Optional[int] = dataclasses.field(default=None)
    failure_code: T.Optional[str] = dataclasses.field(default=None)
    failure_message: T.Optional[str] = dataclasses.field(default=None)
    # fmt: on

    def __post_init__(self):
        if self.s3_prefix is not None:
            if self.s3_prefix.endswith("/") is False:
                self.s3_prefix += "/"

    @classmethod
    def list_imports(
        cls,
        dynamodb_client: "DynamoDBClient",
        table_arn: str,
        page_size: int = 25,
        max_results: int = 1000,
        get_details: bool = False,
    ) -> T.List["ImportJob"]:
        """
        Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/list_imports.html
        """
        imports = list()
        next_token = None
        while 1:
            kwargs = dict(
                TableArn=table_arn,
                MaxResults=page_size,
            )
            if next_token is not None:
                kwargs["NextToken"] = next_token
            res = dynamodb_client.list_imports(**kwargs)
            for dct in res.get("ImportSummaryList", []):
                import_arn = dct["ImportArn"]
                import_status = dct["ImportStatus"]
                if get_details:
                    import_job = cls.describe_import(dynamodb_client, import_arn)
                else:
                    import_job = cls(arn=import_arn, status=import_status)
                imports.append(import_job)
                if len(imports) >= max_results:
                    return imports

            _next_token = res.get("NextToken", "NOT_AVAILABLE")
            if _next_token == "NOT_AVAILABLE":
                break
            else:
                next_token = _next_token
        return imports

    @classmethod
    def from_import_description(cls, desc: dict):
        """
        :param desc: The import description dictionary from
            ``import_table``, ``describe_import`` or ``list_imports``.
        """
        return cls(
            arn=desc["ImportArn"],
            status=desc["ImportStatus"],
            table_arn=desc.get("TableArn"),
            table_id=desc.get("TableId"),
            client_token=desc.get("ClientToken"),
            s3_bucket_owner=desc.get("S3BucketSource", {}).get("S3BucketOwner"),
            s3_bucket=desc.get("S3BucketSource", {}).get("S3Bucket"),
            s3_prefix=desc.get("S3BucketSource", {}).get("S3KeyPrefix"),
            error_count=desc.get("ErrorCount"),
            cloudwatch_log_group_arn=desc.get("CloudWatchLogGroupArn"),
            input_format=desc.get("InputFormat"),
            import_format_options=desc.get("InputFormatOptions"),
            import_compression_type=desc.get("InputCompressionType"),
            table_creation_parameters=desc.get("TableCreationParameters"),
            start_time=desc.get("StartTime"),
            end_time=desc.get("EndTime"),
            processed_size_bytes=desc.get("ProcessedSizeBytes"),
            processed_item_count=desc.get("ProcessedItemCount"),
            imported_item_count=desc.get("ImportedItemCount"),
            failure_code=desc.get("FailureCode"),
            failure_message=desc.get("FailureMessage"),
        )

    @classmethod
    def describe_import(
        cls,
        dynamodb_client: "DynamoDBClient",
        import_arn: str,
    ) -> T.Optional["ImportJob"]:
        """
        Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/describe_import.html
        """
        try:
            res = dynamodb_client.describe_import(ImportArn=import_arn)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ImportNotFoundException":
                return None
            else:
                raise NotImplementedError
        desc = res["ImportTableDescription"]
        return cls.from_import_description(desc)

    def is_in_progress(self) -> bool:
        return self.status == ImportStatusEnum.IN_PROGRESS.value

    def is_completed(self) -> bool:
        return self.status == ImportStatusEnum.COMPLETED.value

    def is_cancelling(self) -> bool:
        return self.status == ImportStatusEnum.CANCELLING.value

    def is_cancelled(self) -> bool:
        return self.status == ImportStatusEnum.CANCELLED.value

    def is_failed(self) -> bool:
        return self.status == ImportStatusEnum.FAILED.value

    @classmethod
    def wait_until_complete(
        cls,
        dynamodb_client: "DynamoDBClient",
        import_arn: str,
        delays: int = 10,
        timeout: int = 900,
        verbose: bool = True,
    ) -> "ImportJob":
        """
        Wait until the DynamoDB import is completed.
        """
        for attempt, elapse in Waiter(
            delays=delays,
            timeout=timeout,
            instant=True,
            verbose=verbose,
        ):
            import_job = cls.describe_import(
                dynamodb_client=dynamodb_client,
                import_arn=import_arn,
            )
            if import_job.is_completed():
                return import_job
            elif import_job.status in [
                ImportStatusEnum.CANCELLING.value,
                ImportStatusEnum.CANCELLED.value,
                ImportStatusEnum.FAILED.value,
            ]:
                raise ValueError(f"Import failed: {import_job.failure_message}")
            else:
                pass


ITEM = "Item"


def write_amazon_ion(
    records: T.Iterable[T.Dict[str, T.Any]],
    s3uri: str,
    s3_client: "S3Client",
) -> "PutObjectOutputTypeDef":
    """
    Write records to S3, so that you can import DynamoDB table from S3 later.

    This function uses the Amazon Ion format and gzip compression.

    Ref:

    - https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/S3DataImport.Format.html#S3DataImport.Requesting.Formats.Ion

    .. note::

        the ``amazon.ion`` python library doesn't work well with the DynamoDB
        import table feature. The ``amazon.ion`` library will encode integer value
        like the standard ``json`` library. ``1`` to ``1``. However, the
        data for import table requires a "dot" after the integer, for example ``1.``.
        Looks like there's no way we can let the ``amazon.ion`` to add the "dot".
        If you want to construct dataset for DynamoDB import table, don't use
        ION, use DynamoDB JSON instead.

    :param records: The list of python dictionary representing the dynamodb item to write to S3.
    :param s3uri: The S3 URI to write the data to.
    :param s3_client: The S3 client to use.
    """
    raise NotImplementedError(
        "amazon.ion library doesn't work well with DynamoDB import table feature. "
        "Use DynamoDB JSON instead."
    )

    # -- this should be the right way to implement it, but it doesn't add the dot
    # -- so that it doesn't work
    # lines = []
    # for record in records:
    #     lines.append(amazon_ion.dumps({ITEM: record}, binary=False))
    # lines.append("")
    # content = gzip.compress("\n".join(lines).encode("utf-8"))

    # -- I also tried this hacky way, but it doesn't work either
    # s = amazon_ion.dumps([{ITEM: record} for record in records], binary=False)
    # content = gzip.compress((s + "\n").encode("utf-8"))

    bucket, key = split_s3_uri(s3uri=s3uri)
    return s3_client.put_object(
        Bucket=bucket,
        Key=key,
        Body=content,
        ContentType="text/plain",
        ContentEncoding="gzip",
    )


def write_dynamodb_json(
    records: T.Iterable[T.Dict[str, T.Any]],
    s3uri: str,
    s3_client: "S3Client",
) -> "PutObjectOutputTypeDef":
    """
    Write records to S3, so that you can import DynamoDB table from S3 later.

    This function uses the DynamoDB JSON format and gzip compression.

    Ref:

    - https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/S3DataImport.Format.html#S3DataImport.Requesting.Formats.DDBJson

    :param records: The list of python dictionary representing the dynamodb item to write to S3.
    :param s3uri: The S3 URI to write the data to.
    :param s3_client: The S3 client to use.
    """
    lines = []
    for record in records:
        lines.append(f'{{"{ITEM}": {dynamodb_json.dumps(record)}}}')
    lines.append("")
    content = gzip.compress("\n".join(lines).encode("utf-8"))
    bucket, key = split_s3_uri(s3uri=s3uri)
    return s3_client.put_object(
        Bucket=bucket,
        Key=key,
        Body=content,
        ContentType="application/json",
        ContentEncoding="gzip",
    )



# -*- coding: utf-8 -*-

from aws_dynamodb_io import api


def test():
    _ = api
    _ = api.T_DNAMODB_JSON
    _ = api.T_DATA
    _ = api.ImportStatusEnum
    _ = api.ImportFormatEnum
    _ = api.ImportJob
    _ = api.ImportJob.list_imports
    _ = api.ImportJob.from_import_description
    _ = api.ImportJob.describe_import
    _ = api.ImportJob.is_in_progress
    _ = api.ImportJob.is_completed
    _ = api.ImportJob.is_cancelling
    _ = api.ImportJob.is_cancelled
    _ = api.ImportJob.is_failed
    _ = api.ImportJob.wait_until_complete
    _ = api.write_amazon_ion
    _ = api.write_dynamodb_json
    _ = api.ManifestSummary
    _ = api.ExportStatusEnum
    _ = api.ExportFormatEnum
    _ = api.DataFile
    _ = api.DataFile.read_dynamodb_json
    _ = api.DataFile.read_amazon_ion
    _ = api.ExportJob
    _ = api.ExportJob.list_exports
    _ = api.ExportJob.describe_export
    _ = api.ExportJob.is_in_progress
    _ = api.ExportJob.is_completed
    _ = api.ExportJob.is_failed
    _ = api.ExportJob.is_dynamodb_json_format
    _ = api.ExportJob.is_ion_format
    _ = api.ExportJob.export_short_id
    _ = api.ExportJob.s3uri_export
    _ = api.ExportJob.s3uri_export_data
    _ = api.ExportJob.s3uri_export_manifest_files
    _ = api.ExportJob.s3uri_export_manifest_summary
    _ = api.ExportJob.get_details
    _ = api.ExportJob.get_manifest_summary
    _ = api.ExportJob.get_data_files
    _ = api.ExportJob.read_dynamodb_json
    _ = api.ExportJob.read_amazon_ion
    _ = api.ExportJob.export_table_to_point_in_time
    _ = api.ExportJob.wait_until_complete


if __name__ == "__main__":
    from aws_dynamodb_io.tests import run_cov_test

    run_cov_test(__file__, "aws_dynamodb_io.api", preview=False)



# -*- coding: utf-8 -*-

"""
DynamoDB export to S3 tool box.

Reference:

- DynamoDB data export to Amazon S3: how it works: https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/S3DataExport.HowItWorks.html
"""

import typing as T
import enum
import json
import gzip
import dataclasses
from datetime import datetime, timezone

import botocore.exceptions

from .importer import amazon_ion, dynamodb_json


from .waiter import Waiter
from .utils import (
    split_s3_uri,
    T_DNAMODB_JSON,
    T_DATA,
)


if T.TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_dynamodb.client import DynamoDBClient
    from amazon.ion.simple_types import IonPyDict


def _parse_time(s: str) -> datetime:
    """
    Parse UTC datetime string into timezone aware datetime object.
    """
    return datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)


@dataclasses.dataclass
class ManifestSummary:
    """
    The ``manifest-summary.json`` file data model.
    """

    table_id: str
    table_arn: str
    s3_bucket: str
    s3_prefix: str
    item_count: int
    output_format: str
    start_time_str: str
    end_time_str: str
    export_time_str: str
    manifest_files_s3_key: str
    billed_size_bytes: int

    @property
    def start_time(self) -> datetime:
        return _parse_time(self.start_time_str)

    @property
    def end_time(self) -> datetime:
        return _parse_time(self.end_time_str)

    @property
    def export_time(self) -> datetime:
        return _parse_time(self.export_time_str)


@dataclasses.dataclass
class DataFile:
    """
    The ``s3://.../AWSDynamoDB/${timestamp}-${random_str]/data/${random_str}.json.gz``
    data file data model.

    :param item_count: number of item in this data file.
    :param md5: md5 hash of this data file.
    :param etag: AWS S3 etag.
    :param s3_bucket: S3 bucket name.
    :param s3_key: S3 key.
    """

    item_count: int
    md5: str
    etag: str
    s3_bucket: str
    s3_key: str
    export_arn: str
    export_format: str

    @property
    def s3_uri(self) -> str:
        """
        The S3 URI of the data file.
        """
        return f"s3://{self.s3_bucket}/{self.s3_key}"

    def read_dynamodb_json(
        self,
        s3_client: "S3Client",
    ) -> T.List[T_DNAMODB_JSON]:
        """
        Read items from the DynamoDB JSON data file.

        Ref: https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/HowItWorks.NamingRulesDataTypes.html

        Example item::

            {
                'key1': {'S': '...'},
                'attr1': {'S': '...'},
                'attr2': {'N': '...'},
                ...
            },
        """
        res = s3_client.get_object(
            Bucket=self.s3_bucket,
            Key=self.s3_key,
        )
        lines = gzip.decompress(res["Body"].read()).decode("utf-8").splitlines()
        return [json.loads(line)["Item"] for line in lines]

    def read_amazon_ion(
        self,
        s3_client: "S3Client",
        ion_loads_kwargs: T.Dict[str, T.Any] = None,
    ) -> T.List["IonPyDict"]:
        """
        Read items from the Amazon ION data file.

        :param s3_client: S3 client for reading data.

        Ref: https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/S3DataExport.Output.html

        Example item::

            {
                'key1': ...,
                'attr1': ...,
                'attr2': ...,
                ...
            },
        """
        res = s3_client.get_object(
            Bucket=self.s3_bucket,
            Key=self.s3_key,
        )
        lines = gzip.decompress(res["Body"].read()).decode("utf-8").splitlines()
        ion_dict_list = list()
        if ion_loads_kwargs is None:
            # use BARE to load native Python object
            ion_loads_kwargs = dict(value_model=amazon_ion.IonPyValueModel.MAY_BE_BARE)
        for line in lines:
            try:
                ion_dict = amazon_ion.loads(line, **ion_loads_kwargs)["Item"]
                ion_dict_list.append(ion_dict)
            except TypeError:
                pass
        return ion_dict_list


class ExportStatusEnum(str, enum.Enum):
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class ExportFormatEnum(str, enum.Enum):
    DYNAMODB_JSON = "DYNAMODB_JSON"
    ION = "ION"


class ExportTypeEnum(str, enum.Enum):
    FULL_EXPORT  = "FULL_EXPORT"
    INCREMENTAL_EXPORT = "INCREMENTAL_EXPORT"


@dataclasses.dataclass
class ExportJob:
    """
    The DynamoDB export table job data model.

    Ref:

    - export_table_to_point_in_time: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/export_table_to_point_in_time.html
    - describe_export: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/describe_export.html
    - list_exports: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/list_exports.html
    - How it works: https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/S3DataExport.HowItWorks.html
    """

    arn: str = dataclasses.field()
    status: str = dataclasses.field()
    start_time: T.Optional[datetime] = dataclasses.field(default=None)
    end_time: T.Optional[datetime] = dataclasses.field(default=None)
    export_time: T.Optional[datetime] = dataclasses.field(default=None)
    table_arn: T.Optional[str] = dataclasses.field(default=None)
    table_id: T.Optional[str] = dataclasses.field(default=None)
    client_token: T.Optional[str] = dataclasses.field(default=None)
    s3_bucket: T.Optional[str] = dataclasses.field(default=None)
    s3_prefix: T.Optional[str] = dataclasses.field(default=None)
    s3_sse_algorithm: T.Optional[str] = dataclasses.field(default=None)
    s3_sse_kms_key_id: T.Optional[str] = dataclasses.field(default=None)
    billed_size_bytes: T.Optional[int] = dataclasses.field(default=None)
    item_count: T.Optional[int] = dataclasses.field(default=None)
    export_format: T.Optional[str] = dataclasses.field(default=None)
    failure_code: T.Optional[str] = dataclasses.field(default=None)
    failure_message: T.Optional[str] = dataclasses.field(default=None)
    export_manifest: T.Optional[str] = dataclasses.field(default=None)
    export_type: T.Optional[str] = dataclasses.field(default=None)
    incremental_export_specification: T.Optional[dict] = dataclasses.field(default=None)

    def __post_init__(self):
        if self.s3_prefix is not None:
            if self.s3_prefix.endswith("/") is False:
                self.s3_prefix += "/"

    @classmethod
    def list_exports(
        cls,
        dynamodb_client: "DynamoDBClient",
        table_arn: str,
        page_size: int = 25,
        max_results: int = 1000,
        get_details: bool = False,
    ) -> T.List["ExportJob"]:
        """
        Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/list_exports.html
        """
        exports = list()
        next_token = None
        while 1:
            kwargs = dict(
                TableArn=table_arn,
                MaxResults=page_size,
            )
            if next_token is not None:
                kwargs["NextToken"] = next_token
            res = dynamodb_client.list_exports(**kwargs)
            for dct in res.get("ExportSummaries", []):
                export_arn = dct["ExportArn"]
                export_status = dct["ExportStatus"]
                if get_details:
                    export_job = cls.describe_export(dynamodb_client, export_arn)
                else:
                    export_job = cls(arn=export_arn, status=export_status)
                exports.append(export_job)
                if len(exports) >= max_results:
                    return exports

            _next_token = res.get("NextToken", "NOT_AVAILABLE")
            if _next_token == "NOT_AVAILABLE":
                break
            else:
                next_token = _next_token
        return exports

    @classmethod
    def from_export_description(cls, desc: dict):
        """
        :param desc: The export description dictionary from
            ``export_table_to_point_in_time``, ``describe_export`` or ``list_exports``.
        """
        return cls(
            arn=desc["ExportArn"],
            status=desc["ExportStatus"],
            start_time=desc.get("StartTime"),
            end_time=desc.get("EndTime"),
            export_time=desc.get("ExportTime"),
            table_arn=desc.get("TableArn"),
            table_id=desc.get("TableId"),
            client_token=desc.get("ClientToken"),
            s3_bucket=desc.get("S3Bucket"),
            s3_prefix=desc.get("S3Prefix"),
            s3_sse_algorithm=desc.get("S3SseAlgorithm"),
            s3_sse_kms_key_id=desc.get("S3SseKmsKeyId"),
            billed_size_bytes=desc.get("BilledSizeBytes"),
            item_count=desc.get("ItemCount"),
            export_format=desc.get("ExportFormat"),
            failure_code=desc.get("FailureCode"),
            failure_message=desc.get("FailureMessage"),
            export_manifest=desc.get("ExportManifest"),
            export_type=desc.get("ExportType"),
            incremental_export_specification=desc.get("IncrementalExportSpecification"),
        )

    @classmethod
    def describe_export(
        cls,
        dynamodb_client: "DynamoDBClient",
        export_arn: str,
    ) -> T.Optional["ExportJob"]:
        """
        Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/describe_export.html
        """
        try:
            res = dynamodb_client.describe_export(ExportArn=export_arn)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ExportNotFoundException":
                return None
            else:
                raise NotImplementedError

        desc = res["ExportDescription"]
        return cls.from_export_description(desc)

    def is_in_progress(self) -> bool:
        return self.status == ExportStatusEnum.IN_PROGRESS.value

    def is_completed(self) -> bool:
        return self.status == ExportStatusEnum.COMPLETED.value

    def is_failed(self) -> bool:
        return self.status == ExportStatusEnum.FAILED.value

    def is_dynamodb_json_format(self) -> bool:
        return self.export_format == ExportFormatEnum.DYNAMODB_JSON.value

    def is_ion_format(self) -> bool:
        return self.export_format == ExportFormatEnum.ION.value

    @property
    def export_short_id(self) -> str:
        """
        The short ID of the export, which is a compound of the export timestamp
        and random string. Example: ``1672531200000-a1b2c3d4``. 1672531200000
        is the timestamp of 2023-01-01 00:00:00
        """
        return self.arn.split("/")[-1]

    @property
    def s3uri_export(self) -> str:
        """
        The S3 folder you specified when you call the
        ``dynamodb_client.export_table_to_point_in_time(...)`` API.

        Example: s3://bucket/prefix/
        """
        return f"s3://{self.s3_bucket}/{self.s3_prefix}"

    @property
    def _s3uri_export_root(self) -> str:
        return f"{self.s3uri_export}AWSDynamoDB/{self.export_short_id}/"

    @property
    def s3uri_export_data(self) -> str:
        """
        Where the export data files are stored.

        Example: s3://bucket/prefix/AWSDynamoDB/1672531200000-a1b2c3d4/data/
        """
        return f"{self._s3uri_export_root}data/"

    @property
    def s3uri_export_manifest_files(self) -> str:
        """
        The S3 location of the manifest files.

        Example: s3://bucket/prefix/AWSDynamoDB/1672531200000-a1b2c3d4/manifest-files.json
        """
        return f"{self._s3uri_export_root}manifest-files.json"

    @property
    def s3uri_export_manifest_summary(self) -> str:
        """
        The S3 location of the manifest summary file.

        Example: s3://bucket/prefix/AWSDynamoDB/1672531200000-a1b2c3d4/manifest-summary.json
        """
        return f"{self._s3uri_export_root}manifest-summary.json"

    def get_details(self, dynamodb_client: "DynamoDBClient"):
        """
        Get the details of the DynamoDB export, refresh it's attributes values.
        """
        export = self.describe_export(
            dynamodb_client=dynamodb_client, export_arn=self.arn
        )
        for field in dataclasses.fields(self.__class__):
            setattr(self, field.name, getattr(export, field.name))
        self.__post_init__()

    def _ensure_details(self, dynamodb_client: "DynamoDBClient"):
        """
        Ensure that the details of the DynamoDB export are available.
        If not, call ``get_details`` to refresh the attributes values.
        """
        if self.s3_bucket is None:
            self.get_details(dynamodb_client=dynamodb_client)

    def get_manifest_summary(
        self,
        dynamodb_client: "DynamoDBClient",
        s3_client: "S3Client",
    ) -> ManifestSummary:
        """
        Get the manifest summary of the DynamoDB export.
        """
        self._ensure_details(dynamodb_client=dynamodb_client)
        bucket, key = split_s3_uri(self.s3uri_export_manifest_summary)
        res = s3_client.get_object(
            Bucket=bucket,
            Key=key,
        )
        data = json.loads(res["Body"].read().decode("utf-8"))
        return ManifestSummary(
            table_id=data["tableId"],
            table_arn=data["tableArn"],
            s3_bucket=data["s3Bucket"],
            s3_prefix=data["s3Prefix"],
            item_count=data["itemCount"],
            output_format=data["outputFormat"],
            start_time_str=data["startTime"],
            end_time_str=data["endTime"],
            export_time_str=data["exportTime"],
            manifest_files_s3_key=data["manifestFilesS3Key"],
            billed_size_bytes=data["billedSizeBytes"],
        )

    def get_data_files(
        self,
        dynamodb_client: "DynamoDBClient",
        s3_client: "S3Client",
    ) -> T.List[DataFile]:
        """
        Get the list of data files of the DynamoDB export.
        """
        self._ensure_details(dynamodb_client=dynamodb_client)
        bucket, key = split_s3_uri(self.s3uri_export_manifest_files)
        res = s3_client.get_object(
            Bucket=bucket,
            Key=key,
        )
        lines = res["Body"].read().decode("utf-8").splitlines()
        data_file_list = list()
        for line in lines:
            data = json.loads(line)
            data_file_list.append(
                DataFile(
                    item_count=data["itemCount"],
                    md5=data["md5Checksum"],
                    etag=data["etag"],
                    s3_bucket=bucket,
                    s3_key=data["dataFileS3Key"],
                    export_arn=self.arn,
                    export_format=self.export_format,
                )
            )
        return data_file_list

    def read_dynamodb_json(
        self,
        dynamodb_client: "DynamoDBClient",
        s3_client: "S3Client",
    ) -> T.Iterable[T_DNAMODB_JSON]:
        """
        Read the items of the DynamoDB export. This is a generator function.
        """
        data_file_list = self.get_data_files(
            dynamodb_client=dynamodb_client,
            s3_client=s3_client,
        )
        for data_file in data_file_list:
            for item in data_file.read_dynamodb_json(s3_client=s3_client):
                yield item

    def read_amazon_ion(
        self,
        dynamodb_client: "DynamoDBClient",
        s3_client: "S3Client",
    ) -> T.Iterable["IonPyDict"]:
        """
        Read the items of the DynamoDB export. This is a generator function.
        """
        data_file_list = self.get_data_files(
            dynamodb_client=dynamodb_client,
            s3_client=s3_client,
        )
        for data_file in data_file_list:
            for item in data_file.read_amazon_ion(
                s3_client=s3_client,
            ):
                yield item

    @classmethod
    def export_table_to_point_in_time(
        cls,
        dynamodb_client: "DynamoDBClient",
        table_arn: str,
        s3_bucket: str,
        s3_prefix: T.Optional[str] = None,
        export_time: T.Optional[datetime] = None,
        s3_bucket_owner: T.Optional[datetime] = None,
        s3_sse_algorithm: T.Optional[datetime] = None,
        s3_sse_kms_key_id: T.Optional[datetime] = None,
        export_format: str = ExportFormatEnum.DYNAMODB_JSON.value,
        export_type: str = ExportTypeEnum.FULL_EXPORT.value,
        client_token: T.Optional[str] = None,
        incremental_export_specification: T.Optional[dict] = None,
    ):
        """
        Export DynamoDB to point-in-time, and return the export object.

        Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/export_table_to_point_in_time.html
        """
        kwargs = dict(
            TableArn=table_arn,
            S3Bucket=s3_bucket,
            S3Prefix=s3_prefix,
            ExportTime=export_time,
            S3BucketOwner=s3_bucket_owner,
            S3SseAlgorithm=s3_sse_algorithm,
            S3SseKmsKeyId=s3_sse_kms_key_id,
            ExportFormat=export_format,
            ExportType=export_type,
            ClientToken=client_token,
            IncrementalExportSpecification=incremental_export_specification,
        )
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        res = dynamodb_client.export_table_to_point_in_time(**kwargs)
        desc = res["ExportDescription"]
        return cls.from_export_description(desc)

    @classmethod
    def wait_until_complete(
        cls,
        dynamodb_client: "DynamoDBClient",
        export_arn: str,
        delays: int = 10,
        timeout: int = 900,
        verbose: bool = True,
    ) -> "ExportJob":
        """
        Wait until the DynamoDB export is completed.
        """
        for attempt, elapse in Waiter(
            delays=delays,
            timeout=timeout,
            instant=True,
            verbose=verbose,
        ):
            export = cls.describe_export(
                dynamodb_client=dynamodb_client,
                export_arn=export_arn,
            )
            if export.is_completed():
                return export
            elif export.is_failed():
                raise ValueError(f"Export failed: {export.failure_message}")
            else:
                pass

    @classmethod
    def from_s3_dir(
        cls,
        s3_client: "S3Client",
        bucket: str,
        prefix: str,
    ):  # pragma: no cover
        """
        Unlike describe_export reading from DynamoDB API, it directly reads the
        export metadata from the S3 folder of a completed export job. The DynamoDB
        export is only available for 35 days after the export is completed. After
        that, you can use this method to read the export from S3 directly.

        :param s3_client: The boto3 S3 client
        :param bucket: The S3 bucket
        :param prefix: It should have a manifest-summary.json file in it.
            Example: "my-dynamodb-export/AWSDynamoDB/01725162280092-940349cc/
        """
        if prefix.endswith("/") is False:
            prefix = prefix + "/"
        # check if it a valid export folder
        s3_client.head_object(Bucket=bucket, Key=f"{prefix}manifest-files.md5")
        s3_client.head_object(Bucket=bucket, Key=f"{prefix}manifest-files.json")
        s3_client.head_object(Bucket=bucket, Key=f"{prefix}manifest-summary.md5")
        res = s3_client.get_object(Bucket=bucket, Key=f"{prefix}manifest-summary.json")
        data = json.loads(res["Body"].read().decode("utf-8"))
        return cls(
            arn=data["exportArn"],
            status=ExportStatusEnum.COMPLETED.value,
            start_time=datetime.strptime(data.get("startTime"), "%Y-%m-%dT%H:%M:%S.%fZ"),
            end_time=datetime.strptime(data.get("endTime"), "%Y-%m-%dT%H:%M:%S.%fZ"),
            export_time=datetime.strptime(data.get("exportTime"), "%Y-%m-%dT%H:%M:%S.%fZ"),
            table_arn=data.get("tableArn"),
            table_id=data.get("tableId"),
            s3_bucket=data.get("s3Bucket"),
            s3_prefix=data.get("s3Prefix"),
            s3_sse_algorithm=data.get("s3SseAlgorithm"),
            s3_sse_kms_key_id=data.get("s3SseKmsKeyId"),
            billed_size_bytes=data.get("billedSizeBytes"),
            item_count=data.get("itemCount"),
            export_format=data.get("outputFormat"),
            export_manifest=data.get("manifestFilesS3Key"),
            export_type=data.get("exportType"),
            incremental_export_specification=data.get("IncrementalExportSpecification"),
        )



# -*- coding: utf-8 -*-

import typing as T


def split_s3_uri(s3uri: str) -> T.Tuple[str, str]:
    """
    Parse S3 URI into bucket and key.

    Example::

        >>> split_s3_uri("s3://my-bucket/folder/file.txt")
        ('my-bucket', 'folder/file.txt')
    """
    parts = s3uri.split("/", 3)
    bucket = parts[2]
    key = parts[3]
    return bucket, key


# Represent a DynamoDB JSON data
# {"key": {"S": "encoded_value"}}
T_DNAMODB_JSON = T.Dict[str, T.Dict[str, T.Any]]

# Represent a native Python dictionary data
T_DATA = T.Dict[str, T.Any]



# -*- coding: utf-8 -*-

import typing as T
import sys
import time
import itertools

__version__ = "0.2.1"


class Waiter:
    """
    Simple retry / polling with progressing status. Usage, it is common to check
    if a long-running job is done every X seconds and timeout in Y seconds.
    This class allow you to customize the polling interval and timeout,.

    Example:

    .. code-block:: python

        print("before waiter")

        for attempt, elapse in Waiter(
            delays=1,
            timeout=10,
            verbose=True,
        ):
            # check if should jump out of the polling loop
            if elapse >= 5:
                print("")
                break

        print("after waiter")


    :param delays: delay between each check
    :param timeout: timeout in seconds
    :param instant: if True, then the first check is instant
    :param indent: indent level for logging
    :param verbose: whether to print log

    .. versionchanged:: 0.2.1

        add instant parameter to make the first check instant
    """

    def __init__(
        self,
        delays: T.Union[int, float],
        timeout: T.Union[int, float],
        instant: bool = False,
        indent: int = 0,
        verbose: bool = True,
    ):
        self._delays = delays
        self.delays = itertools.repeat(delays)
        self.timeout = timeout
        self.instant = instant
        self.tab = " " * indent
        self.verbose = verbose

    def __iter__(self):
        k = 1
        start = time.time()
        end = start + self.timeout

        if self.instant:
            initial_attempt = 1
        else:
            initial_attempt = 0

        if self.verbose:
            sys.stdout.write(
                f"start waiter, polling every {self._delays} seconds, "
                f"timeout in {self.timeout} seconds.\n"
            )
            sys.stdout.flush()
            sys.stdout.write(
                f"\r{self.tab}on {initial_attempt} th attempt, "
                f"elapsed 0 seconds, "
                f"remain {self.timeout} seconds ..."
            )
            sys.stdout.flush()

        if self.instant:
            k += 1
            yield 1, 0

        for attempt, delay in enumerate(self.delays, k):
            now = time.time()
            remaining = end - now
            if remaining < 0:
                raise TimeoutError(f"timed out in {self.timeout} seconds!")
            else:
                time.sleep(min(delay, remaining))
                elapsed = int(now - start + delay)
                if self.verbose:  # pragma: no cover
                    sys.stdout.write(
                        f"\r{self.tab}on {attempt} th attempt, "
                        f"elapsed {elapsed} seconds, "
                        f"remain {self.timeout - elapsed} seconds ..."
                    )
                    sys.stdout.flush()
                yield attempt, int(elapsed)


# -*- coding: utf-8 -*-


def test():
    data = {"id": 1, "name": "Alice"}
    try:
        from aws_dynamodb_io.importer import amazon_ion

        amazon_ion.dumps(data)
    except ImportError:
        pass

    try:
        from aws_dynamodb_io.importer import dynamodb_json

        dynamodb_json.dumps(data)
    except ImportError:
        pass


if __name__ == "__main__":
    from aws_dynamodb_io.tests import run_cov_test

    run_cov_test(__file__, "aws_dynamodb_io.importer", preview=False)



# -*- coding: utf-8 -*-


amazon_ion_error = ImportError(
    "You don't have amazon-ion installed. "
    "Please install it to use this feature. "
    "See https://pypi.org/project/amazon.ion/ for more details"
)


class FakeAmazonIon:  # pragma: no cover
    def dumps(self, *args, **kwargs):
        raise amazon_ion_error

    def loads(self, *args, **kwargs):
        raise amazon_ion_error

    @property
    def IonPyValueModel(self):
        raise amazon_ion_error


dynamodb_json_error = ImportError(
    "You don't have dynamodb-json installed. "
    "Please install it to use this feature. "
    "See https://pypi.org/project/dynamodb-json/ for more details"
)


class FakeDynamoDBJson:  # pragma: no cover
    def loads(self, *args, **kwargs):
        raise dynamodb_json_error

    def dumps(self, *args, **kwargs):
        raise dynamodb_json_error


try:  # pragma: no cover
    import amazon.ion.simpleion as amazon_ion
    from amazon.ion.simple_types import IonPyDict

    has_amazon_ion = True
except ImportError:  # pragma: no cover
    amazon_ion = FakeAmazonIon()
    has_amazon_ion = False

try:  # pragma: no cover
    from dynamodb_json import json_util as dynamodb_json

    has_dynamodb_json = True
except ImportError:  # pragma: no cover
    dynamodb_json = FakeDynamoDBJson()
    has_dynamodb_json = False
